
; Flop's Pi Calculator
;
; Copyright 2001 Benjamin Moody
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version.
; This program is distributed in the hope that it will be useful, but WITHOUT
; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
; details.
; You should have received a copy of the GNU General Public License along with
; this program; if not, write to the Free Software Foundation, Inc., 675 Mass
; Ave, Cambridge, MA 02139, USA.

.nolist
#include "ion.inc"
.list

kRight  = 1
kLeft   = 2
kUp     = 3
kDown   = 4
kEnter  = 5
kClear  = 9
kDel    = 10
kQuit   = 64
kMode   = 69
k0      = 142
k1      = 143
k2      = 144

#ifdef TI83
    .org progStart
#else
    .org progStart-2
    .db $bb,$6d
#endif

Start:
    ld (QuitSaveSP),sp
    res textwrite,(iy+sgrflags)
    set texterasebelow,(iy+textflags)
    res indiconly,(iy+indicflags)
    res indicrun,(iy+indicflags)
    res apptextsave,(iy+appflags)
#ifdef TI83P
    in a,(2)
    and $80
    jr z,NotSE
    rlca
    out ($20),a
NotSE:
#endif
    ld hl,ProgName
    ld de,OP1
    ld bc,9
    ldir
    bcall(_chkfindsym)
    jr c,NormalStartup
    inc de
    inc de
    ex de,hl
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld (MP_Pi),hl
    ld (NumDigits),de
    bcall(_clrlcdfull)
    call DispTitle
    ld hl,2
    ld (curRow),hl
    ld hl,NewPromptStr
    bcall(_puts)
NewLoop:
    bcall(_getkey)
    cp kQuit
    jp z,QuitSimple
    cp kMode
    jp z,QuitSimple
    cp kClear
    jp z,QuitSimple
    cp k1
    jr z,NormalStartup
    cp k2
    jr nz,NewLoop
    jp ViewZero
NormalStartup:
    bcall(_clrlcdfull)
    call DispTitle
    ld hl,2
    ld (curRow),hl
    ld hl,PromptStr
    bcall(_puts)
EntryClear:
    ld hl,0
    ld (NumDigits),hl
EntryLoop:
    ld hl,4*256+5
    ld (curRow),hl
    ld hl,(NumDigits)
    bcall(_disphl)
EntryKeyLoop:
    bcall(_getkey)
    cp kEnter
    jr z,EntryDone
    cp kDel
    jr z,EntryDel
    cp kLeft
    jr z,EntryDel
    cp kClear
    jr z,EntryClear
    cp kMode
    jp z,QuitSimple
    cp kQuit
    jp z,QuitSimple
    sub k0
    jr c,EntryKeyLoop
    cp 10
    jr nc,EntryKeyLoop
    ld hl,(NumDigits)
    ld b,h
    ld c,l
    add hl,hl
    jr c,EntryMax
    add hl,hl
    jr c,EntryMax
    add hl,bc
    jr c,EntryMax
    add hl,hl
    jr c,EntryMax
    ld c,a
    ld b,0
    add hl,bc
    ld (NumDigits),hl
    jr EntryLoop
EntryMax:
    ld hl,$7fff
    ld (NumDigits),hl
    jr EntryLoop
EntryDel:
    ld hl,(NumDigits)
    bcall(_divhlby10)
    ld (NumDigits),hl
    jr EntryLoop
EntryDone:
    ld hl,(NumDigits)
    ld a,h
    or l
    jr z,EntryLoop
    srl h
    rr l
    inc hl
    ld (NumDigits),hl
    bcall(_memfree)
    ; Do we have enough?
    ex de,hl
    ld hl,(NumDigits)
    add hl,hl
    jr c,NotEnoughMem   ; in this case, you have *way* too many digits.
    add hl,hl
    jr c,NotEnoughMem
#ifdef TI83
    ld bc,9
#else
    ld bc,12
#endif
    add hl,bc
    jr c,NotEnoughMem
    sbc hl,de
    jr c,AdequateMem
NotEnoughMem:
    ld hl,7
    ld (curRow),hl
    ld hl,MemStr
    bcall(_puts)
    jp EntryLoop
AdequateMem:
    add hl,de
    push hl
    ld hl,ProgName
    ld de,OP1
    ld bc,9
    ldir
    bcall(_chkfindsym)
    jr c,ProgramDoesntExist
#ifdef TI83
    bcall(_delvar)
#else
    bcall(_delvararc)
#endif
ProgramDoesntExist:
    pop hl
    inc hl  ; Save the number of digits
    inc hl
#ifdef TI83
    bcall(_createprog)
    ld (hl),6
#else
    bcall(_createprotprog)
#endif
    ld hl,(NumDigits)
    ex de,hl
    inc hl
    inc hl
    ld de,(NumDigits)
    ld (hl),e
    inc hl
    ld (hl),d
    inc hl
    ld (MP_Pi),hl
    add hl,de
    ld (MP_ACot),hl
    add hl,de
    ld (MP_ACot_Term),hl
    add hl,de
    ld (MP_ACot_Power),hl

    bcall(_clrlcdfull)
    call DispTitle
    ld hl,1
    ld (curRow),hl
    ld hl,CalculatingStr
    bcall(_puts)

; OK! Let's start calculating!
    ld hl,7
    ld (curRow),hl
    ld hl,CalcIterationStr
    bcall(_puts)

    di

    call CheckOnQuit    ; *** ON-key check ***
    
    call DispN0
    ld hl,1*256+2
    ld (curRow),hl
    ld hl,Calc18Str
    bcall(_puts)
    ld a,18
    call ACotMP

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot)
    ld e,48
    call MulMPByInt
    ld hl,(MP_ACot)
    ld de,(MP_Pi)
    ld bc,(NumDigits)
    ldir

    call CheckOnQuit    ; *** ON-key check ***

    call DispN0
    ld hl,1*256+3
    ld (curRow),hl
    ld hl,Calc57Str
    bcall(_puts)
    ld a,57
    call ACotMP

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot)
    ld e,32
    call MulMPByInt
    ld de,(MP_ACot)
    ld hl,(MP_Pi)
    call AddMP

    call CheckOnQuit    ; *** ON-key check ***

    call DispN0
    ld hl,1*256+4
    ld (curRow),hl
    ld hl,Calc239Str
    bcall(_puts)
    ld a,239
    call ACotMP

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot)
    ld e,20
    call MulMPByInt
;   ld hl,(MP_ACot)
;   ld de,(MP_Pi)
;   ld bc,(NumDigits)
;   ldir
    ld de,(MP_ACot)
    ld hl,(MP_Pi)
    call SubMP

;    ld hl,(MP_Pi)
;    call ZeroMP
;    ld hl,(MP_Pi)
;    ld (hl),1
;    ld de,239*239
;    call DivMPByInt

    ei

ViewZero:
    ld hl,0
ViewSet:
    ld (DisplayPos),hl
ViewLoop:
    bcall(_clrlcdfull)
    ld hl,3
    ld (curRow),hl
    ld a,$C4
    bcall(_putmap)
    ld de,(DisplayPos)
    ld a,d
    or e
    jr z,ViewNoUpArrow
    ld hl,0
    ld (penCol),hl
    ld a,$1E
    bcall(_vputmap)
    ld de,(DisplayPos)
    ld hl,2*256+8
    ld (penCol),hl
    ld hl,(MP_Pi)
    jr ViewUpArrow
ViewNoUpArrow:
    ld hl,2*256+8
    ld (penCol),hl
    ld hl,FirstStr
    bcall(_vputs)
    ld hl,(MP_Pi)
    ld de,(DisplayPos)
    inc de
ViewUpArrow:
    call DisplayMP_X
    jr nc,ViewNoDownArrow
    ld hl,56*256
    ld (penCol),hl
    ld a,$1F
    bcall(_vputmap)
ViewNoDownArrow:
ViewKeyLoop:
    bcall(_getkey)
    cp kMode
    jp z,QuitSimple
    cp kQuit
    jp z,QuitSimple
    cp kClear
    jp z,QuitSimple
    cp kUp
    jr z,ViewUp
    cp kDown
    jr z,ViewDown
    cp kLeft
    jr z,ViewPUp
    cp kRight
    jr z,ViewPDn
    jr ViewKeyLoop
ViewUp:
    ld hl,(DisplayPos)
    ld bc,-10
    add hl,bc
    jr nc,ViewKeyLoop
    jp ViewSet
ViewDown:
    ld hl,(DisplayPos)
    ld bc,10
    add hl,bc
    ld bc,(NumDigits)
    or a
    sbc hl,bc
    jr nc,ViewKeyLoop
    add hl,bc
    jp ViewSet
ViewPUp:
    ld hl,(DisplayPos)
    ld bc,-90
    add hl,bc
    jp nc,ViewZero
    jp ViewSet
ViewPDn:
    ld hl,(DisplayPos)
    ld bc,90
    add hl,bc
    ld bc,(NumDigits)
    or a
    sbc hl,bc
    jr nc,ViewKeyLoop
    add hl,bc
    jp ViewSet

#ifdef TI83
onport = 3
#else
onport = 4
#endif

CheckOnQuit:
    ; Call this periodically, let the user stop the calculation
    push af
    in a,(onport)
    and 8
    jr z,OnKey
    pop af
    ret
OnKey:
    in a,(onport)
    and 8
    jr z,OnKey

    push de
    push bc
    push hl
    push ix

    ld hl,1
    ld (curRow),hl
    ld hl,PauseStr
    bcall(_puts)

    ei
    bcall(_getkey)
    di

    cp kMode
    jr z,Quit
    cp kQuit
    jr z,Quit
    cp kClear
    jr z,Quit

    ld hl,1
    ld (curRow),hl
    ld hl,CalculatingStr
    bcall(_puts)

    pop ix
    pop hl
    pop bc
    pop de
    pop af
    ret

Quit:
    ld sp,(QuitSaveSP)
QuitSimple:
    bcall(_clrlcdfull)
    ld hl,0
    ld (curRow),hl
    set apptextsave,(iy+appflags)
    ei
    ret

DispTitle:
    ld hl,10
    ld (penCol),hl
    set textinverse,(iy+textflags)
    ld a,6
    bcall(_vputmap)
    ld hl,TitleStr
    bcall(_vputs)
    ld a,6
    bcall(_vputmap)
    res textinverse,(iy+textflags)
    ret

DispN0:
    ld hl,2*256+7
    ld (curRow),hl
    ld hl,0
    bcall(_disphl)
    ret

AddMP:
    ; HL->MP number to add into
    ; DE->MP number to add
    ; note: doesn't check for final carry...
    ld bc,(NumDigits)
    add hl,bc
    ex de,hl
    add hl,bc
    inc bc
AddMPLoopNoCarry:

    call CheckOnQuit    ; *** ON-key check ***

    dec hl
    dec de
    dec bc
    ld a,b
    or c
    ret z
    ld a,(de)
    add a,(hl)
    ld (de),a
    sub 100
    jr c,AddMPLoopNoCarry
AddMPLoopCarry:
    ld (de),a   ; we carry, get rid of the 100 in this digit

    call CheckOnQuit    ; *** ON-key check ***

    dec hl
    dec de
    dec bc
    ld a,b
    or c
    ret z
    ld a,(de)
    scf         ; add in our carry
    adc a,(hl)
    ld (de),a
    sub 100
    jr c,AddMPLoopNoCarry
    jr AddMPLoopCarry

SubMP:
    ; HL->MP number to subtract into
    ; DE->MP number to subtract
    ; note: doesn't check for final carry...
    ld bc,(NumDigits)
    add hl,bc
    ex de,hl
    add hl,bc
SubMPLoopNoCarry:

    call CheckOnQuit    ; *** ON-key check ***

    dec hl
    dec de
    dec bc
    ld a,b
    or c
    ret z
    ld a,(de)
    sub (hl)
    ld (de),a
    jr nc,SubMPLoopNoCarry
SubMPLoopCarry:
    add a,100
    ld (de),a   ; we carry, add the 100 to this digit

    call CheckOnQuit    ; *** ON-key check ***

    dec hl
    dec de
    dec bc
    ld a,b
    or c
    ret z
    ld a,(de)
    scf
    sbc a,(hl)
    ld (de),a
    jr nc,SubMPLoopNoCarry
    jr SubMPLoopCarry

MulMPByInt:
    ; HL->MP number to multiply
    ; E = integer to multiply by
    ld bc,(NumDigits)
    ld ixh,b
    ld ixl,c
    add hl,bc
    ld c,0  ; C = carry
MulMPByIntLoop:

    call CheckOnQuit    ; *** ON-key check ***

    dec hl
    ld a,ixh
    or ixl
    ret z
    ld a,(hl)
    push hl
    ld h,a
    call HtimesE
    add hl,bc   ; add in the carry
    ld c,100
    call HLdivC
    ld c,l      ; new carry
    pop hl
    ld (hl),a   ; remainder
    dec ix
    jr MulMPByIntLoop

DivMPByInt:
    ; HL->MP number to divide
    ; DE = number to divide by
    ld (Divisor),de
    ld bc,(NumDigits)
    push bc
    ld c,(hl)
    push hl
    xor a
    ld b,a
    jr DivMPByIntFirst
DivMPByIntLoop:
    push bc

    call CheckOnQuit    ; *** ON-key check ***

    ; Remainder = IX
    ld a,(hl)
    push hl
    ld b,ixh
    ld c,ixl
    ld h,b
    ld l,c
    ld ixh,a
    xor a
    ld e,a
    ; Multiply AHL by 100
    add hl,hl
    adc a,a
    add hl,hl
    adc a,a
    add hl,bc
    adc a,e
    ld c,l
    ld b,h
    ld e,a
    add hl,hl
    adc a,a
    add hl,hl
    adc a,a
    add hl,bc
    adc a,e
    add hl,hl
    adc a,a
    add hl,hl
    adc a,a
    ; add in the next digit
    ld c,ixh
    ld b,0
    ld e,b
    add hl,bc
    adc a,e
    ld b,h
    ld c,l
    ld de,(Divisor)
DivMPByIntFirst:
    call ABCdivDE ; divide into one digit... this is the most time-consuming part
    ld a,c
    ex de,hl
    ld ixh,d
    ld ixl,e
    ex de,hl
    pop hl
    ld (hl),a
    inc hl
    pop bc
    dec bc
    ld a,b
    or c
    jr nz,DivMPByIntLoop
    ret

ACotMP:
    ; A = integer to take inverse cotangent
    ld (ACot_X),a
    ld h,a
    ld e,a
    call HtimesE
    ld (ACot_Xsquared),hl
    
    ld hl,(MP_ACot)
    call ZeroMP
    ld hl,(MP_ACot)
    ld (hl),1
    ld a,(ACot_X)
    ld e,a
    ld d,0
    call DivMPByInt

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot)
    ld de,(MP_ACot_Power)
    ld bc,(NumDigits)
    ldir
    
    ld hl,3
    ld (ACot_N),hl

ACotMPLoop:
    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot_Power)
    ld de,(ACot_Xsquared)
    call DivMPByInt

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot_Power)
    ld de,(MP_ACot_Term)
    ld bc,(NumDigits)
    ldir

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot_Term)
    ld de,(ACot_N)
    call DivMPByInt

    call CheckOnQuit    ; *** ON-key check ***

    ld hl,(MP_ACot_Term)
    call Cp0_MP
    ret z

    call CheckOnQuit    ; *** ON-key check ***
    
    ld hl,(ACot_N)
    inc hl
#ifndef NO_DISPLAY_N
;    ld a,l
;    and $0f
;    jr nz,ACotNoWrite

    push hl
    srl h
    rr l
    ld de,2*256+7
    ld (curRow),de
    bcall(_disphl)
    pop hl
ACotNoWrite:
#endif
    bit 1,l
    inc hl
    ld (ACot_N),hl

    call CheckOnQuit    ; *** ON-key check ***
    
    ld hl,(MP_ACot)
    ld de,(MP_ACot_Term)

    jr z,ACotMP_Sub
    call AddMP
    jr ACotMPLoop
ACotMP_Sub:
    call SubMP
    jr ACotMPLoop

ZeroMP:
    ; HL->MP number to set to zero
    ld (hl),0
    ld d,h
    ld e,l
    ld bc,(NumDigits)
    dec bc
    inc de
    ldir
    ret

Cp0_MP:
    ; HL->MP number to check for zeroness
    ld bc,(NumDigits)
Cp0_MPLoop:
    xor a
    cp (hl)
    ret nz
    inc hl
    dec bc
    ld a,b
    or c
    jr nz,Cp0_MPLoop
    ret

DisplayMP:
    ; HL->MP number to display
    ; Returns NC if whole number fit
    ld de,0
DisplayMP_X:
    ; HL->MP number to display
    ; DE = where to start
    add hl,de
    push hl
    ld hl,(NumDigits)
    sbc hl,de
    ld b,h
    ld c,l
    pop hl
DisplayMPLoop:
    ld a,(hl)
    ld d,'0'-1
DisplayMPDigitLoop:
    inc d
    sub 10
    jr nc,DisplayMPDigitLoop
    add a,10+'0'
    push af
    ld a,d
    call VPutC
    pop de
    ret c
    ld a,d
    call VPutC
    ret c
    inc hl
    dec bc
    ld a,b
    or c
    jr nz,DisplayMPLoop
    ret


HtimesE:
    ; HL = product
    ld l,0
    ld d,l
    ld b,7
    sla h
    jr nc,HtimesEloop
    ld l,e
HtimesEloop:
    add hl,hl
    jr nc,HtimesE_1
    add hl,de
HtimesE_1:
    djnz HtimesEloop
    ret


HLdivC:
    ; HL = quotient, A = remainder
    xor a
    ld b,16
HLdivCloop:
    add hl,hl
    rla
    cp c
    jr c,HLdivCnext
    sub c
    inc l
HLdivCnext:
    djnz HLdivCloop
    ret

ABCdivDE:
    ; ABC = quotient, HL = remainder
    ld hl,0
    ld ixh,24
ABCdivDEloop:
    sla c
    rl b
    rl a
    adc hl,hl
    jr c,ABCdivDE_ROver
    sbc hl,de
    jr c,ABCdivDE_NoBit
    inc c
    dec ixh
    jr nz,ABCdivDEloop
    ret
ABCdivDE_NoBit:
    add hl,de
    dec ixh
    jr nz,ABCdivDEloop
    ret
ABCdivDE_ROver:
    or a
    sbc hl,de
    inc c
    dec ixh
    jr nz,ABCdivDEloop
    ret

VPutC:
    ld d,a
    ld a,(penCol)
    cp 86
    jr c,VPutC_OK
    ld a,8
    ld (penCol),a
    ld a,(penRow)
    add a,6
    cp 58
    ccf
    ret c
    ld (penRow),a
VPutC_OK:
    ld a,d
    bcall(_vputmap)
    ret

TitleStr:
    .db "Flop's "
    .db $C4
    .db " Calculator",0

PromptStr:
    .db "How much ",$C4," would"
    .db "    you like?",0

; Well? How much C4 *would* you like? :)

MemStr:
    .db "Not enough RAM!",0

ProgName:
    .db 6,"pi",0

NewPromptStr:
    .db "1: Recalculate  "
    .db "2: View Old",0

CalculatingStr:
    .db "Calculating...",0
Calc18Str:
    .db "cot",$11," 18",0
Calc57Str:
    .db "cot",$11," 57",0
Calc239Str:
    .db "cot",$11," 239",0
CalcIterationStr:
    .db 1,"=",0

PauseStr:
    .db "   - Paused - ",0

FirstStr:
    .db "3 . ",0

QuitSaveSP:     .dw 0
Divisor:        .dw 0
DisplayPos:     .dw 0
NumDigits:      .dw 0
MP_Pi:          .dw 0
MP_ACot:        .dw 0
MP_ACot_Power:  .dw 0
MP_ACot_Term:   .dw 0
ACot_X:         .db 0
ACot_Xsquared:  .dw 0
ACot_N:         .dw 0

.end
END
